/**
 * Copyright (c) 2004 Logos GmbH Rostock (www.logos-ib.de).
 * All rights reserved.
 *
 * THIS SOFTWARE IS  PROVIDED  "AS IS"  AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF  MERCHANTABILITY  AND  FITNESS  FOR  A PARTICULAR  PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL  LOGOS GMBH  OR ITS CONTRIBUTORS BE
 * LIABLE FOR ANY  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES  (INCL., BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES;  LOSS OF USE, DATA, OR  PROFITS;
 * OR BUSINESS INTERRUPTION)  HOWEVER  CAUSED  AND  ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,  OR TORT (INCL.
 * NEGLIGENCE OR OTHERWISE)  ARISING  IN  ANY  WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package de.bsvrz.dua.bastband.bastband.dvs.funclib;

/**
 * History:
 * 2007-07-26 [KS]  Änderung Konfigurationsanfrage Objekte nach PIDs: weil
 *                  Datenzeitstempel vor Objekt-Gültigkeitsbereich liegen kann,
 *                  werden jetzt alle Objektversionen angefragt, die seit
 *                  dem Anfragebeginn gültig waren bzw. es jetzt sind
 * 2007-07-13 [KS]  NEU: Suchfunktion nach jüngstem Datensatz (Datenzeitstempel)
 *                  UPDATE: Suchfunktion nach jüngstem Datensatz (Archivzeit)
 *
 * $Log: ArchiveRequestHelper.java,v $
 * Revision 1.6  2007/08/21 07:47:03  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.5  2007/07/30 11:18:50  Strey
 * Änderung Archivanfrage über Objekt-PIDs: angefragt wird einschließlich der aktuell gültigen Objektversion
 *
 * Revision 1.4  2007/07/25 13:02:05  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.3  2007/07/13 13:46:53  Strey
 * Anpassung Archivzugriff mit Suchfunktion nach jüngstem Datensatz nach Datenzeitstempel bzw. Archivzeit.
 *
 * Revision 1.2  2007/06/20 12:36:13  Trier
 * <Kein Kommentar eingegeben>
 *
 * Revision 1.1.1.1  2007/06/20 09:56:26  Trier
 * dvs alles
 *
 * Revision 1.1  2006/11/13 09:24:05  Trier
 * <Kein Kommentar eingegeben>
 *
 *
 */
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKind;
import de.bsvrz.dav.daf.main.archive.ArchiveDataKindCombination;
import de.bsvrz.dav.daf.main.archive.ArchiveDataQueryResult;
import de.bsvrz.dav.daf.main.archive.ArchiveDataSpecification;
import de.bsvrz.dav.daf.main.archive.ArchiveDataStream;
import de.bsvrz.dav.daf.main.archive.ArchiveOrder;
import de.bsvrz.dav.daf.main.archive.ArchiveQueryPriority;
import de.bsvrz.dav.daf.main.archive.ArchiveRequestOption;
import de.bsvrz.dav.daf.main.archive.ArchiveTimeSpecification;
import de.bsvrz.dav.daf.main.archive.TimingType;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * <p>
 * stellt Funktionen zum Abfragen der Archivs bereit, die die Daten f&uuml;r
 * alle Versionen, der durch die &uuml;bergebenen Pids pr&auml;sentierenen
 * Objekte, bestimmt.
 * </p>
 * <p>
 * Einige Funktionen filtern die Daten nach dem Zeitstempel so, dass nur die
 * Datens&auml;tze zur&uuml;ckgegeben werden, die innerhalb der angegebenen
 * Zeitspanne liegen.
 * </p>
 *
 */
abstract public class ArchiveRequestHelper {

	private static final Debug LOGGER = Debug.getLogger();

	private ArchiveRequestHelper() {
	}

	private static final ArsDataComparator ARS_DATA_COMPARATOR = new ArsDataComparator();

	private static final ArchiveDataKindCombination DATA_KIND_COMBINATION = new ArchiveDataKindCombination(
			ArchiveDataKind.ONLINE, ArchiveDataKind.ONLINE_DELAYED, ArchiveDataKind.REQUESTED,
			ArchiveDataKind.REQUESTED_DELAYED);

	/**
	 * gibt das Ergebnis f&uuml;r die ArchivAnfrage nach den SystemObjekten mit
	 * der Pid im angegebenen Zeitraum <b>ungefiltert</b> zur&uuml;ck.
	 *
	 * @param dav
	 *            ClientDavInterface
	 * @param soPid
	 *            String
	 * @param atgPid
	 *            String
	 * @param aspPid
	 *            String
	 * @param von
	 *            long
	 * @param bis
	 *            long
	 * @return ArchiveDataQueryResult
	 * @throws IOException
	 *             Fehler beim Lesen der Daten vom Archiv
	 */
	public static ArchiveDataQueryResult getArchiveDataQueryResult(final ClientDavInterface dav, final String soPid,
			final String atgPid, final String aspPid, final long von, final long bis) throws IOException {
		final ArchiveTimeSpecification timeSpec = new ArchiveTimeSpecification(TimingType.DATA_TIME, false, von, bis);
		ArchiveRequestHelper.LOGGER.finer("getArchiveDataQueryResult(): timeSpec=", timeSpec);
		// (im Zeitraum) gültige Versionen des Systemobjekts ermitteln
		final Collection<SystemObject> sos = dav.getDataModel().getObjects(soPid, 0, System.currentTimeMillis());
		final List<ArchiveDataSpecification> specs = new ArrayList<>(sos.size());

		for (final SystemObject so : sos) {
			final AttributeGroup atg = ArchiveRequestHelper.getAttributeGroup(so, atgPid);
			final Aspect asp = ArchiveRequestHelper.getAspect(atg, aspPid);
			final DataDescription dataDescr = new DataDescription(atg, asp, (short) 0);
			ArchiveRequestHelper.LOGGER.finer("getArchiveDataQueryResult(): so=" + so + ", dataDescr=" + dataDescr);
			final ArchiveDataSpecification dataSpec = new ArchiveDataSpecification(timeSpec,
					ArchiveRequestHelper.DATA_KIND_COMBINATION, ArchiveOrder.BY_INDEX, ArchiveRequestOption.NORMAL,
					dataDescr, so);

			specs.add(dataSpec);
		}
		return dav.getArchive().request(ArchiveQueryPriority.MEDIUM, specs);
	}

	/**
	 * gibt das Ergebnis f&uuml;r die ArchivAnfrage nach den SystemObjekten mit
	 * der Pid im angegebenen Zeitraum <b>ungefiltert</b> als Stream
	 * zur&uuml;ck.
	 *
	 * @param dav
	 *            ClientDavInterface
	 * @param soPid
	 *            String
	 * @param atgPid
	 *            String
	 * @param aspPid
	 *            String
	 * @param von
	 *            long
	 * @param bis
	 *            long
	 * @return ArchiveDataStream
	 * @throws IOException
	 *             Fehler beim Lesen der Daten vom Archiv
	 * @throws InterruptedException
	 *             das Einlesen der Daten wurde unerwartet unterbrochen
	 */
	public static ArchiveDataStream getStream(final ClientDavInterface dav, final String soPid, final String atgPid,
			final String aspPid, final long von, final long bis) throws IOException, InterruptedException {

		final ArchiveDataQueryResult arsResult = ArchiveRequestHelper.getArchiveDataQueryResult(dav, soPid, atgPid,
				aspPid, von, bis);
		return new MergedArchiveDataStream(arsResult);
	}

	/**
	 * gibt die sortierten ArchivDaten im angegebenen Bereich
	 * (einschlie&szlig;lich der Grenzen) <b>sortiert</b> und <b>gefiltert</b>
	 * zur&uuml;ck.
	 *
	 * @param dav
	 *            ClientDavInterface
	 * @param soPid
	 *            String
	 * @param atgPid
	 *            String
	 * @param aspPid
	 *            String
	 * @param von
	 *            long
	 * @param bis
	 *            long
	 * @throws IOException
	 *             wenn Fehler beim Lesen aufgetreten sind.
	 * @return Set
	 */
	public static Set<ArchiveData> getArchivData(final ClientDavInterface dav, final String soPid, final String atgPid,
			final String aspPid, final long von, final long bis) throws IOException {

		final TreeSet<ArchiveData> set = new TreeSet<>(ArchiveRequestHelper.ARS_DATA_COMPARATOR);
		final ArchiveDataQueryResult arsResult = ArchiveRequestHelper.getArchiveDataQueryResult(dav, soPid, atgPid,
				aspPid, von, bis);
		try {
			final ArchiveDataStream[] streams = arsResult.getStreams();
			ArchiveRequestHelper.LOGGER.finer("getArchiveData(): streams.length=" + streams.length);
			for (final ArchiveDataStream stream : streams) {
				ArchiveData arsData = stream.take();
				while (arsData != null) {
					if ((arsData.getDataTime() <= bis) && (arsData.getDataTime() >= von)) {
						set.add(arsData);
					}
					arsData = stream.take();
				}
			}
		} catch (final IOException ex) {
			throw ex;
		} catch (final IllegalStateException ex) {
			throw new IOException(ex.getMessage());
		} catch (final InterruptedException ex) {
			throw new IOException(ex.getMessage());
		}
		return set;
	}

	/**
	 * bestimmt die AttributeGroup mit der Pid f&uuml;r das angegebene
	 * SystemObject.
	 *
	 * @param so
	 *            SystemObject
	 * @param atgPid
	 *            String
	 * @return AttributeGroup
	 */
	public static final AttributeGroup getAttributeGroup(final SystemObject so, final String atgPid) {
		final List<AttributeGroup> atgList = so.getType().getAttributeGroups();
		for (final AttributeGroup atg : atgList) {
			if (atg.getPid().equals(atgPid)) {
				return atg;
			}
		}
		return null;
	}

	/**
	 * bestimmt den Aspect mit der Pid f&uuml;r die angegebene AttributeGroup.
	 *
	 * @param atg
	 *            AttributeGroup
	 * @param aspPid
	 *            String
	 * @return Aspect
	 */
	public static final Aspect getAspect(final AttributeGroup atg, final String aspPid) {
		final Collection<Aspect> aspList = atg.getAspects();
		for (final Aspect asp : aspList) {
			if (asp.getPid().equals(aspPid)) {
				return asp;
			}
		}
		return null;
	}

	/**
	 * <p>
	 * Bestimmung des zuletzt archivierten Datensatzes aus den im Archiv
	 * vorliegenden Daten für das gegebene Systemobjekt.
	 * </p>
	 * <p>
	 * Da das Archivsystem als Zustandsarchiv und nicht als Datenhaltung
	 * ausgelegt ist, ist eine verlässliche Suche nur mit Brute-Force Methoden
	 * möglich. Dazu werden alle Änderungen bis zu einer maximalen Zeitspanne ab
	 * aktuellem Zeitpunkt abgefragt und der zu letzt archivierte Datensatz
	 * gesucht.
	 *
	 * Falls die Archivanfrage keine (gültigen) Ergebnisse enthält, wird der
	 * Wert <code>null</code> zurückgegeben.
	 *
	 * Ausgewertet werden alle Online, nachgelieferten und nachgeforderten Werte
	 * von allen im Anfragezeitraum gültigen Versionen des Objekts.
	 * </p>
	 *
	 * @param dav
	 *            Datenverteilerverbindung
	 * @param so
	 *            das anzufragenden Objekt
	 * @param descr
	 *            Datenbeschreibung der gesuchten Daten
	 * @param maxHistory
	 *            long der maximale Zeitbereich in ms, bis zu dem vom aktuellen
	 *            zeitpunkt aus nach einem Datensatz gesucht werden soll
	 * @return der Archivdatensatz mit dem jüngsten Archivzeitstempel im
	 *         Anfragezeitraum
	 * @throws IllegalStateException
	 *             Falls die Archivanfrage nicht erfolgreich war und keine
	 *             Ergebnisdatenströme bestimmt werden können.
	 */
	public static ArchiveData getLastArchiveDataByArchiveTime(final ClientDavInterface dav, final SystemObject so,
			final DataDescription descr, final long maxHistory) {
		ArchiveData result = null;
		final long systemTime = System.currentTimeMillis();
		final long queryTo = systemTime;
		final long queryFrom = queryTo - maxHistory;
		final ArchiveTimeSpecification timeSpec = new ArchiveTimeSpecification(TimingType.DATA_TIME, false, queryFrom,
				queryTo);
		try {
			// im Zeitraum gültige Versionen des Systemobjekts ermitteln und
			// anfragen
			final Collection<SystemObject> queryObjects = dav.getDataModel().getObjects(so.getPid(), queryFrom,
					systemTime);
			for (final SystemObject queryObject : queryObjects) {
				final ArchiveDataSpecification dataSpec = new ArchiveDataSpecification(timeSpec,
						ArchiveRequestHelper.DATA_KIND_COMBINATION, ArchiveOrder.BY_INDEX, ArchiveRequestOption.NORMAL,
						descr, queryObject);
				final ArchiveDataQueryResult arsQueryResult = dav.getArchive().request(ArchiveQueryPriority.MEDIUM,
						dataSpec);
				for (final ArchiveDataStream stream : arsQueryResult.getStreams()) {
					ArchiveData arsData = stream.take();
					while (arsData != null) {
						// Auswertung Archivantwort / Suche jüngsten
						// Archivzeitstempel
						if (arsData.getData() != null) {
							if ((result == null) || (arsData.getDataTime() > result.getDataTime())) {
								result = arsData;
							}
						}
						arsData = stream.take();
					}
				}
			}
		} catch (final Exception ex) {
			ArchiveRequestHelper.LOGGER.warning("getLastArchiveDataByArchiveTime(): Fehler bei Archivanfrage", ex);
			throw new IllegalStateException(ex);
		}

		ArchiveRequestHelper.LOGGER.finer(" getLastArchiveDataByArchiveTime(" + so.getPid() + ':'
				+ descr.getAttributeGroup().getPid() + ':' + descr.getAspect().getPid() + ") " + "==> "
				+ ((result == null) ? "null" : "data " + new Date(result.getDataTime())));

		return result;
	}

	/**
	 * <p>
	 * Bestimmung des jüngsten Datensatzes aus den im Archiv vorliegenden Daten
	 * für das gegebene Systemobjekt.
	 * </p>
	 * <p>
	 * Da das Archivsystem als Zustandsarchiv und nicht als Datenhaltung
	 * ausgelegt ist, ist eine verlässliche Suche nur mit Brute-Force Methoden
	 * möglich. Dazu werden alle Änderungen bis zu einer maximalen Zeitspanne ab
	 * aktuellem Zeitpunkt abgefragt und in den Antwortdaten der Datensatz mit
	 * dem jüngsten Datenzeitstempel gesucht. Daten die außerhalb des
	 * angefragten Bereichs liegen werden ignoriert.
	 * </p>
	 * <p>
	 * Falls die Archivanfrage keine (gültigen) Ergebnisse enthält, wird der
	 * Wert <code>null</code> zurückgegeben.
	 * </p>
	 * <p>
	 * Ausgewertet werden alle Online, nachgelieferten und nachgeforderten Werte
	 * von allen im Anfragezeitraum gültigen Versionen des Objekts.
	 * </p>
	 *
	 * @param connection
	 *            {@link ClientDavInterface} Datenverteilerverbindung
	 * @param object
	 *            {@link SystemObject} das anzufragenden Objekt
	 * @param descr
	 *            {@link DataDescription} Datenbeschreibung der gesuchten Daten
	 * @param fromTime
	 *            long Beginn Suchbereich
	 * @param toTime
	 *            long Ende Suchbereich
	 * @return {@link ArchiveData} der Archivdatensatz mit dem jüngsten
	 *         Datenzeitstempel im Anfragezeitraum
	 * @throws IllegalStateException
	 *             Falls die Archivanfrage nicht erfolgreich war und keine
	 *             Ergebnisdatenströme bestimmt werden können.
	 */
	public static ArchiveData getLastArchiveDataByDataTime(final ClientDavInterface connection,
			final SystemObject object, final DataDescription descr, final long fromTime, final long toTime)
					throws java.lang.IllegalStateException {
		final long minHistory = IntervalHelper.TAG_IN_MS * 32; // mindestens 1
		// Monat
		// durchsuchen
		ArchiveData result = null;
		final long systemTime = System.currentTimeMillis();
		final long queryTo = Math.max(toTime, systemTime);
		final long queryFrom = Math.min(fromTime, systemTime - minHistory);

		final ArchiveTimeSpecification timeSpec = new ArchiveTimeSpecification(TimingType.DATA_TIME, false, queryFrom,
				queryTo);
		final ArchiveDataKindCombination dataKinds = new ArchiveDataKindCombination(ArchiveDataKind.ONLINE,
				ArchiveDataKind.ONLINE_DELAYED, ArchiveDataKind.REQUESTED, ArchiveDataKind.REQUESTED_DELAYED);

		try {
			// im Zeitraum gültige Versionen des Systemobjekts ermitteln und
			// anfragen
			//
			// Weil Daten auch vor dem Beginn der Gültigkeit des Objekts
			// eingespielt
			// werden können, muss mindestens bis zum aktuellen Zeitpunkt nach
			// gültigen Objekten gesucht werden !
			//
			final Collection<SystemObject> queryObjects = connection.getDataModel().getObjects(object.getPid(),
					fromTime, systemTime);
			for (final SystemObject queryObject : queryObjects) {
				final ArchiveDataSpecification dataSpec = new ArchiveDataSpecification(timeSpec, dataKinds,
						ArchiveOrder.BY_INDEX, ArchiveRequestOption.NORMAL, descr, queryObject);
				final ArchiveDataQueryResult arsQueryResult = connection.getArchive()
						.request(ArchiveQueryPriority.MEDIUM, dataSpec);
				for (final ArchiveDataStream stream : arsQueryResult.getStreams()) {
					ArchiveData arsData = stream.take();
					while (arsData != null) {
						// Auswertung Archivantwort / Suche jüngsten Datensatz
						if ((arsData.getData() != null) && (arsData.getDataTime() >= fromTime)
								&& (arsData.getDataTime() <= toTime)) {
							if ((result == null) || (arsData.getDataTime() > result.getDataTime())) {
								result = arsData;
							}
						}
						arsData = stream.take();
					}
				}
			}
		} catch (final Exception ex) {
			ArchiveRequestHelper.LOGGER.warning("getLastArchiveDataByDataTime(): Fehler bei Archivanfrage", ex);
			throw new IllegalStateException(ex);
		}

		ArchiveRequestHelper.LOGGER.finer("getLastArchiveDataByDataTime(" + object.getPid() + ':'
				+ descr.getAttributeGroup().getPid() + ':' + descr.getAspect().getPid() + ") " + "==> "
				+ ((result == null) ? "null" : "data " + new Date(result.getDataTime())));

		return result;
	}

	/**
	 * Comparator zum Sortieren von ArchiveData nach DataTime, DataIndex.
	 *
	 * @author unbekannt
	 * @version 1.0
	 */
	static class ArsDataComparator implements Comparator<ArchiveData> {

		@Override
		public int compare(final ArchiveData o1, final ArchiveData o2) {
			if (o1.getDataTime() == o2.getDataTime()) {
				return (o1.getDataIndex() < o2.getDataIndex()) ? -1 : 1;
			}
			return (o1.getDataTime() < o2.getDataTime()) ? -1 : 1;
		}

		@Override
		public boolean equals(final Object obj) {
			return (this.getClass() == obj.getClass());
		}
	}

	/**
	 * ArchiveDataStream, der mehrere Streams zu einem Stream
	 * zusammenf&uuml;hrt.
	 *
	 * @author unbekannt
	 * @version 1.0
	 */
	static class MergedArchiveDataStream implements ArchiveDataStream {

		ArchiveDataStream[] _Streams;
		int actStreamNr = 0;

		MergedArchiveDataStream(final ArchiveDataQueryResult queryResult)
				throws IllegalStateException, InterruptedException {
			_Streams = queryResult.getStreams();
		}

		@Override
		public ArchiveDataSpecification getDataSpecification() {
			return _Streams[actStreamNr].getDataSpecification();
		}

		@Override
		public ArchiveData take() throws InterruptedException, IOException, IllegalStateException {
			if (actStreamNr < _Streams.length) {
				final ArchiveData arsData = _Streams[actStreamNr].take();
				if (arsData == null) {
					actStreamNr++;
					return this.take();
				}
				return arsData;
			}
			return null;
		}

		@Override
		public void abort() {
			for (final ArchiveDataStream stream : _Streams) {
				if (stream != null) {
					stream.abort();
				}
			}
		}

	}

}
